﻿namespace TabsStudioDisambiguator
{
    public class Disambiguator : TabsStudioExt.ITabsStudioAddin
    {
        public void OnConnection(TabsStudioExt.ITabsStudioEngine engine_, EnvDTE80.DTE2 dte)
        {
            engine = engine_;
            isPreviewTabProperty = new AddinsCommon.IsPreviewTabProperty();
            engine.TabCreated += OnTabCreated;
        }

        public void OnDisconnection()
        {
            engine.TabCreated -= OnTabCreated;
        }

        private void OnTabCreated(object sender, TabsStudioExt.TabEventArgs e)
        {
            string title = GetTabName(e.Tab);
            System.Collections.Generic.List<TabInfo> tabsWithSameTitle =
                new System.Collections.Generic.List<TabInfo>();
            foreach(TabsStudioExt.Tab tab in e.Tabs.TabList)
            {
                if (!isPreviewTabProperty.IsPreviewTab(tab.TabItem) && string.Compare(GetTabName(tab), title, true) == 0)
                {
                    TabInfo info = GetTabInfo(tab);
                    if (info.folder != null || info.project != null || info.directory != null)
                        tabsWithSameTitle.Add(info);
                }
            }
            if (tabsWithSameTitle.Count > 1)
            {
                DisambiguationType type = SelectDisambiguation(tabsWithSameTitle);
                foreach (TabInfo tab in tabsWithSameTitle)
                {
                    RemoveDisambiguatorTag(tab.tab);
                    AddDisambiguatorTag(tab.tab, tab.GetParent(type));
                }
            }
        }

        private void AddDisambiguatorTag(TabsStudioExt.Tab tab, string parent)
        {
            tab.TitleManager.Transforms.Add(
                new TabsStudioExt.TitleTransform("^.*$", parent + "/" + "$_", disambiguatorTag));
            tab.TitleManager.ApplyTransforms();
        }

        private void RemoveDisambiguatorTag(TabsStudioExt.Tab tab)
        {
            foreach (TabsStudioExt.TitleTransform t in tab.TitleManager.Transforms)
            {
                if (t.Id == disambiguatorTag)
                {
                    tab.TitleManager.Transforms.Remove(t);
                    tab.TitleManager.ApplyTransforms();
                    return;
                }
            }
        }

        private DisambiguationType SelectDisambiguation(System.Collections.Generic.List<TabInfo> tabs)
        {
            UniqueStringsCounter projectCounter = new UniqueStringsCounter();
            UniqueStringsCounter folderCounter = new UniqueStringsCounter();
            UniqueStringsCounter directoryCounter = new UniqueStringsCounter();
            foreach (TabInfo i in tabs)
            {
                projectCounter.Add(i.project);
                folderCounter.Add(i.folder);
                directoryCounter.Add(i.directory);
            }
            uint maxCount = System.Math.Max(directoryCounter.GetCount(),
                System.Math.Max(folderCounter.GetCount(), projectCounter.GetCount()));
            return (projectCounter.GetCount() == maxCount || folderCounter.GetCount() == maxCount) ?
                        (folderCounter.GetCount() > projectCounter.GetCount() ||
                        (folderCounter.GetCount() == 1 && projectCounter.GetCount() == 1) ? 
                            DisambiguationType.Folder : 
                            DisambiguationType.Project) 
                        :
                DisambiguationType.Directory;
        }

        private string GetTabName(TabsStudioExt.Tab tab)
        {
            return GetTilteWithoutDisambiguator(tab);
        }

        private string GetTilteWithoutDisambiguator(TabsStudioExt.Tab tab)
        {
            string result = tab.TitleManager.UntransformedSplitTitle != null ? tab.TitleManager.UntransformedSplitTitle : tab.TitleManager.UntransformedTitle;
            foreach (TabsStudioExt.TitleTransform t in tab.TitleManager.Transforms)
            {
                if (!t.Id.Equals(disambiguatorTag))
                    result = System.Text.RegularExpressions.Regex.Replace(result, t.PatternRegex, t.ReplacementRegex);
            }
            return result;
        }

        private TabInfo GetTabInfo(TabsStudioExt.Tab tab)
        {
            TabInfo result = new TabInfo(tab);
            EnvDTE.Window w = tab.TabExtensions[0].Window;
            try
            {
                if (w.Document != null && w.Document.ProjectItem != null && w.Document.ProjectItem.Collection != null)
                {
                    EnvDTE.ProjectItem pi = w.Document.ProjectItem.Collection.Parent as EnvDTE.ProjectItem;
                    while (pi != null && pi.Kind == EnvDTE.Constants.vsProjectItemKindPhysicalFile)
                        pi = pi.Collection.Parent as EnvDTE.ProjectItem;
                    if (pi != null)
                        result.folder = pi.Name;
                }
            }
            catch (System.Exception)
            {
            }

            try
            {
                if (w.Document != null && w.Document.ProjectItem != null && 
                    w.Document.ProjectItem.ContainingProject != null)
                {
                    string projectName = w.Document.ProjectItem.ContainingProject.Name;
                    if (System.IO.Path.IsPathRooted(projectName))
                        projectName = System.IO.Path.GetFileName(System.IO.Path.GetDirectoryName(projectName));
                    result.project = projectName;
                }
            }
            catch (System.Exception)
            {
            }

            try
            {
                if (w.Document != null && w.Document.FullName != null)
                    result.directory = System.IO.Path.GetFileName(System.IO.Path.GetDirectoryName(w.Document.FullName));
            }
            catch (System.Exception)
            {
            }
            
            return result;
        }

        private TabsStudioExt.ITabsStudioEngine engine;
        private AddinsCommon.IsPreviewTabProperty isPreviewTabProperty;
        private static string disambiguatorTag = "[Disambiguator]";
    }
}
